# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from ..internal.special import cdf_bounds, erf, erfinv, ppf_bounds_cont
from ..internal.optimization import optimize_ml
from ..internal.distribution_helper import eps, all_not_none
from .distributions import Continuous


class TruncatedNormal(Continuous):
    r"""
    TruncatedNormal distribution.

    The pdf of this distribution is

    .. math::

       f(x;\mu ,\sigma ,a,b)={\frac {\phi ({\frac {x-\mu }{\sigma }})}{
            \sigma \left(\Phi ({\frac {b-\mu }{\sigma }})-\Phi ({\frac {a-\mu }{\sigma }})\right)}}

    .. plot::
        :context: close-figs


        from preliz import TruncatedNormal, style
        style.use('preliz-doc')
        mus = [0.,  0., 0.]
        sigmas = [3.,5.,7.]
        lowers = [-3, -5, -5]
        uppers = [7, 5, 4]
        for mu, sigma, lower, upper in zip(mus, sigmas,lowers,uppers):
            TruncatedNormal(mu, sigma, lower, upper).plot_pdf(support=(-10,10))

    ========  ==========================================
    Support   :math:`x \in [a, b]`
    Mean      :math:`\mu +{\frac {\phi (\alpha )-\phi (\beta )}{Z}}\sigma`
    Variance  .. math::
                  \sigma ^{2}\left[1+{\frac {\alpha \phi (\alpha )-\beta \phi (\beta )}{Z}}-
                  \left({\frac {\phi (\alpha )-\phi (\beta )}{Z}}\right)^{2}\right]
    ========  ==========================================

    Parameters
    ----------
    mu : float
        Mean.
    sigma : float
        Standard deviation (sigma > 0)
    lower: float
        Lower limit.
    upper: float
        Upper limit (upper > lower).
    """

    def __init__(self, mu=None, sigma=None, lower=None, upper=None):
        super().__init__()
        self._parametrization(mu, sigma, lower, upper)

    def _parametrization(self, mu=None, sigma=None, lower=None, upper=None):
        self.mu = mu
        self.sigma = sigma
        self.lower = lower
        self.upper = upper
        self.params = (self.mu, self.sigma, self.lower, self.upper)
        self.param_names = ("mu", "sigma", "lower", "upper")
        self.params_support = (
            (-np.inf, np.inf),
            (eps, np.inf),
            (-np.inf, np.inf),
            (-np.inf, np.inf),
        )
        if lower is None:
            self.lower = -np.inf
        if upper is None:
            self.upper = np.inf
        self.support = (self.lower, self.upper)
        if all_not_none(mu, sigma, lower, upper):
            self._update(mu, sigma, lower, upper)

    def _update(self, mu, sigma, lower=None, upper=None):
        if lower is not None:
            self.lower = np.float64(lower)
        if upper is not None:
            self.upper = np.float64(upper)

        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.params = (self.mu, self.sigma, self.lower, self.upper)
        self.support = (self.lower, self.upper)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.sigma, self.lower, self.upper))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma, self.lower, self.upper)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma, self.lower, self.upper)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma, self.lower, self.upper)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma, self.lower, self.upper)

    def entropy(self):
        return nb_entropy(self.mu, self.sigma, self.lower, self.upper)

    def mean(self):
        alpha = (self.lower - self.mu) / self.sigma
        beta = (self.upper - self.mu) / self.sigma
        z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
        return (
            self.mu
            + (
                (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * alpha**2))
                - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * beta**2))
            )
            / z_val
            * self.sigma
        )

    def median(self):
        alpha = (self.lower - self.mu) / self.sigma
        beta = (self.upper - self.mu) / self.sigma
        inv_phi = 2**0.5 * erfinv(
            2 * ((0.5 * (1 + erf(alpha / 2**0.5)) + 0.5 * (1 + erf(beta / 2**0.5))) / 2) - 1
        )
        return self.mu + inv_phi * self.sigma

    def var(self):
        alpha = (self.lower - self.mu) / self.sigma
        beta = (self.upper - self.mu) / self.sigma
        z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
        # Handle for -np.inf or np.inf
        psi_alpha = (0, 0) if alpha == -np.inf else (1, alpha)
        psi_beta = (0, 0) if beta == np.inf else (1, beta)
        return self.sigma**2 * (
            1
            - (
                psi_beta[1] * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                - psi_alpha[1] * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
            )
            / z_val
            - (
                (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                / z_val
            )
            ** 2
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        alpha = (self.lower - self.mu) / self.sigma
        beta = (self.upper - self.mu) / self.sigma
        z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
        # Handle for -np.inf or np.inf
        psi_alpha = (0, 0) if alpha == -np.inf else (1, alpha)
        psi_beta = (0, 0) if beta == np.inf else (1, beta)
        numerator = (
            (
                (psi_alpha[1] ** 2 - 1)
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                * psi_alpha[0]
                - (psi_beta[1] ** 2 - 1)
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                * psi_beta[0]
            )
            / z_val
            - 3
            * (
                psi_alpha[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                * psi_alpha[0]
                - psi_beta[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                * psi_beta[0]
            )
            * (
                (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
            )
            / z_val**2
            + 2
            * (
                (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                / z_val
            )
            ** 3
        )
        denominator = (
            1
            + (
                psi_alpha[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                * psi_alpha[0]
                - psi_beta[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                * psi_beta[0]
            )
            / z_val
            - (
                (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                / z_val
            )
            ** 2
        ) ** (3 / 2)
        return numerator / denominator

    def kurtosis(self):
        alpha = (self.lower - self.mu) / self.sigma
        beta = (self.upper - self.mu) / self.sigma
        z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
        # Handle for -np.inf or np.inf
        psi_alpha = (0, 0) if alpha == -np.inf else (1, alpha)
        psi_beta = (0, 0) if beta == np.inf else (1, beta)

        numerator = (
            (
                12
                * (
                    psi_alpha[1]
                    * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                    * psi_alpha[0]
                    - psi_beta[1]
                    * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                    * psi_beta[0]
                )
                * (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                ** 2
                / z_val**3
            )
            - (
                4
                * (
                    (psi_alpha[1] ** 2 - 1)
                    * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                    * psi_alpha[0]
                    - (psi_beta[1] ** 2 - 1)
                    * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                    * psi_beta[0]
                )
                * (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                / z_val**2
            )
            - (
                3
                * (
                    (
                        psi_alpha[1]
                        * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                        * psi_alpha[0]
                        - psi_beta[1]
                        * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                        * psi_beta[0]
                    )
                    / z_val
                )
                ** 2
            )
            - (
                6
                * (
                    (
                        (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                        - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                    )
                    / z_val
                )
                ** 4
            )
            + (
                (psi_alpha[1] ** 3 - 3 * psi_alpha[1])
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                * psi_alpha[0]
                - (psi_beta[1] ** 3 - 3 * psi_beta[1])
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                * psi_beta[0]
            )
            / z_val
        )

        denominator = (
            1
            + (
                psi_alpha[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2))
                * psi_alpha[0]
                - psi_beta[1]
                * (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2))
                * psi_beta[0]
            )
            / z_val
            - (
                (
                    (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[0]
                    - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[0]
                )
                / z_val
            )
            ** 2
        ) ** 2

        return numerator / denominator

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(0, 1, size)
        return nb_rvs(random_samples, self.mu, self.sigma, self.lower, self.upper)

    def _fit_moments(self, mean, sigma):
        # Assume gaussian
        self._update(mean, sigma)

    def _fit_mle(self, sample):
        self._update(None, None, np.min(sample), np.max(sample))
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, mu, sigma, lower, upper):
    xi = (x - mu) / sigma
    alpha = (lower - mu) / sigma
    beta = (upper - mu) / sigma
    z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
    prob = (0.5 * (1 + erf(xi / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))) / z_val
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma, lower, upper):
    alpha = (lower - mu) / sigma
    beta = (upper - mu) / sigma
    inv_phi = 2**0.5 * erfinv(
        2
        * (
            0.5 * (1 + erf(alpha / 2**0.5))
            + q * (0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5)))
        )
        - 1
    )
    return ppf_bounds_cont(inv_phi * sigma + mu, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(mu, sigma, lower, upper):
    alpha = (lower - mu) / sigma
    beta = (upper - mu) / sigma
    z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
    # Handle for -np.inf or np.inf
    psi_alpha = (0, 0) if alpha == -np.inf else (1, alpha)
    psi_beta = (0, 0) if beta == np.inf else (1, beta)
    return np.log((2 * np.pi * np.e) ** 0.5 * sigma * z_val) + (
        (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_alpha[1] ** 2)) * psi_alpha[1] * psi_alpha[0]
        - (1 / (2 * np.pi) ** 0.5 * np.exp(-0.5 * psi_beta[1] ** 2)) * psi_beta[1] * psi_beta[0]
    ) / (2 * z_val)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, sigma, lower, upper):
    if x < lower or x > upper:
        return -np.inf
    else:
        xi = (x - mu) / sigma
        alpha = (lower - mu) / sigma
        beta = (upper - mu) / sigma
        z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
        logphi = np.log(1 / (2 * np.pi) ** 0.5) - xi**2 / 2
        return logphi - (np.log(sigma) + np.log(z_val))


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma, lower, upper):
    return -(nb_logpdf(x, mu, sigma, lower, upper)).sum()


@nb.njit(cache=True)
def nb_rvs(random_samples, mu, sigma, lower, upper):
    alpha = (lower - mu) / sigma
    beta = (upper - mu) / sigma
    z_val = 0.5 * (1 + erf(beta / 2**0.5)) - 0.5 * (1 + erf(alpha / 2**0.5))
    inv_phi = 2**0.5 * erfinv(
        2 * (0.5 * (1 + erf(alpha / 2**0.5)) + random_samples * z_val) - 1
    )
    return inv_phi * sigma + mu



from os import path
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor


def run_notebook(notebook):
    current_dir = path.dirname(path.realpath(__file__))
    file_path = path.join(current_dir, notebook)

    with open(file_path, encoding="utf-8") as inb_f:
        inb = nbformat.read(inb_f, as_version=4)

    exec_pre = ExecutePreprocessor(timeout=600, kernel_name="python3")

    try:
        exec_pre.preprocess(inb, {"metadata": {"path": current_dir}})
    except Exception as e:
        raise RuntimeError(f"Notebook execution failed: {e}") from e

    for cell in inb.cells:
        if cell.cell_type == "code" and "outputs" in cell:
            for output in cell.outputs:
                try:
                    if "FAILED" in output.text:
                        raise RuntimeError(f"Error in notebook cell:\n{output.text}")
                except AttributeError:
                    pass



import warnings
import numpy as np
from preliz.distributions import Dirichlet, Beta
from preliz.internal.optimization import optimize_dirichlet_mode
from preliz.internal.rcparams import rcParams


def dirichlet_mode(mode, mass=None, bound=0.01, plot=None, plot_kwargs=None, ax=None):
    """
    Returns a Dirichlet distribution where the marginals have the specified mode
    and mass and their masses lie within the range mode ± bound.

    Parameters
    ----------
    mode : list
        Mode of the Dirichlet distribution.
    mass : float
        Probability mass between within mode +- bounds. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.
    bound : float
        Defines upper and lower bounds for the mass as mode +- bound. Defaults to 0.01.
    plot : bool
        Whether to plot the distribution. Defaults to True.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()``.
    ax : matplotlib axes

    Returns
    -------
    ax : matplotlib axes
    dist : Preliz Dirichlet distribution.
        Dirichlet distribution with fitted parameters alpha for the given mass and intervals.

    References
    ----------
    Adapted from  Evans et al. (2017) see https://doi.org/10.3390/e19100564
    """
    if mass is None:
        mass = rcParams["stats.ci_prob"]

    if plot is None:
        plot = rcParams["plots.show_plot"]

    if not 0 < mass <= 1:
        raise ValueError("mass should be larger than 0 and smaller or equal to 1")

    if not all(i > 0 for i in mode):
        raise ValueError("mode should be larger than 0")

    if not abs(sum(mode) - 1) < 0.0001:
        warnings.warn("The mode should sum to 1, normalising mode to sum to 1")
        sum_mode = sum(mode)
        mode = [i / sum_mode for i in mode]

    if plot_kwargs is None:
        plot_kwargs = {}

    lower_bounds = np.clip(np.array(mode) - bound, 0, 1)
    target_mass = (1 - mass) / 2
    _dist = Beta()

    _, alpha = optimize_dirichlet_mode(lower_bounds, mode, target_mass, _dist)

    alpha_np = np.array(alpha)
    calculated_mode = (alpha_np - 1) / (alpha_np.sum() - len(alpha_np))

    if np.any((np.array(mode) - calculated_mode) > 0.01):
        warnings.warn(
            f"The requested mode {mode} is different from the calculated mode {calculated_mode}."
        )

    dirichlet_distribution = Dirichlet(alpha)

    if plot:
        ax = dirichlet_distribution.plot_pdf(**plot_kwargs, ax=ax)

    return ax, dirichlet_distribution



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_cont, mean_sample, xlog1py


class Exponential(Continuous):
    r"""
    Exponential Distribution

    The pdf of this distribution is

    .. math::

        f(x \mid \lambda) = \lambda \exp\left\{ -\lambda x \right\}

    .. plot::
        :context: close-figs


        from preliz import Exponential, style
        style.use('preliz-doc')
        for lam in [0.5,  2.]:
            Exponential(lam).plot_pdf(support=(0,5))


    ========  ============================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\dfrac{1}{\lambda}`
    Variance  :math:`\dfrac{1}{\lambda^2}`
    ========  ============================

    Exponential distribution has 2 alternative parametrizations. In terms of lambda (rate)
    or in terms of beta (scale).

    The link between the two alternatives is given by:

    .. math::

        \beta = \dfrac{1}{\lambda}

    Parameters
    ----------
    lam : float
        Rate or inverse scale (lam > 0).
    beta : float
        Scale (beta > 0).
    """

    def __init__(self, lam=None, beta=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(lam, beta)

    def _parametrization(self, lam=None, beta=None):
        if all_not_none(lam, beta):
            raise ValueError("Incompatible parametrization. Either use 'lam' or 'beta'.")

        self.param_names = ("lam",)
        self.params_support = ((eps, np.inf),)

        if beta is not None:
            lam = 1 / beta
            self.param_names = ("beta",)

        self.lam = lam
        self.beta = beta
        if self.lam is not None:
            self._update(self.lam)

    def _update(self, lam):
        self.lam = np.float64(lam)
        self.beta = 1 / self.lam

        if self.param_names[0] == "lam":
            self.params = (self.lam,)
        elif self.param_names[0] == "beta":
            self.params = (self.beta,)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.lam))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.lam)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.beta)

    def logpdf(self, x):
        return nb_logpdf(x, self.lam)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.lam)

    def entropy(self):
        return nb_entropy(self.beta)

    def median(self):
        return np.log(2) * self.beta

    def mean(self):
        return self.beta

    def mode(self):
        return np.zeros_like(self.beta)

    def std(self):
        return self.beta

    def var(self):
        return self.beta**2

    def skewness(self):
        return 2

    def kurtosis(self):
        return 6

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.exponential(self.beta, size)

    def _fit_moments(self, mean, sigma=None):  # pylint: disable=unused-argument
        lam = 1 / mean
        self._update(lam)

    def _fit_mle(self, sample):
        mean = mean_sample(sample)
        self._update(1 / mean)


@nb.njit(cache=True)
def nb_cdf(x, lam):
    x_lam = lam * x
    return cdf_bounds(1 - np.exp(-x_lam), x, 0, np.inf)


@nb.njit(cache=True)
def nb_ppf(q, beta):
    return ppf_bounds_cont(-xlog1py(beta, -q), q, 0, np.inf)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, lam):
    if x < 0:
        return -np.inf
    else:
        return np.log(lam) - lam * x


@nb.njit(cache=True)
def nb_neg_logpdf(x, lam):
    return (-nb_logpdf(x, lam)).sum()


@nb.njit(cache=True)
def nb_entropy(beta):
    return 1 + np.log(beta)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import erf, erfinv, zeta  # pylint: disable=no-name-in-module

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import erf, erfinv, ppf_bounds_cont
from ..internal.optimization import optimize_ml


class Moyal(Continuous):
    r"""
    Moyal distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu,\sigma) =
            \frac{1}{\sqrt{2\pi}\sigma}e^{-\frac{1}{2}\left(z + e^{-z}\right)},

    where

    .. math::

       z = \frac{x-\mu}{\sigma}

    .. plot::
        :context: close-figs


        from preliz import Moyal, style
        style.use('preliz-doc')
        mus = [-1., 0., 4.]
        sigmas = [2., 1., 4.]
        for mu, sigma in zip(mus, sigmas):
            Moyal(mu, sigma).plot_pdf(support=(-10,20))

    ========  ==============================================================
    Support   :math:`x \in (-\infty, \infty)`
    Mean      :math:`\mu + \sigma\left(\gamma + \log 2\right)`, where
              :math:`\gamma` is the Euler-Mascheroni constant
    Variance  :math:`\frac{\pi^{2}}{2}\sigma^{2}`
    ========  ==============================================================

    Parameters
    ----------
    mu : float
        Location parameter.
    sigma : float
        Scale parameter (sigma > 0).
    """

    def __init__(self, mu=None, sigma=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, sigma)

    def _parametrization(self, mu=None, sigma=None):
        self.mu = mu
        self.sigma = sigma
        self.params = (self.mu, self.sigma)
        self.param_names = ("mu", "sigma")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        if all_not_none(mu, sigma):
            self._update(self.mu, self.sigma)

    def _update(self, mu, sigma):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.params = (self.mu, self.sigma)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        return self.mu + self.sigma * (np.euler_gamma + np.log(2))

    def mode(self):
        return self.mu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.sigma**2 * (np.pi**2) / 2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 28 * np.sqrt(2) * zeta(3) / np.pi**3

    def kurtosis(self):
        return 4

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return self.ppf(random_state.random(size))

    def _fit_moments(self, mean, sigma):
        sigma = sigma / np.pi * 2**0.5
        mu = mean - sigma * (np.euler_gamma + np.log(2))
        self._update(mu, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, mu, sigma):
    z_val = (x - mu) / sigma
    return 1 - erf(np.exp(-z_val / 2) * (2**-0.5))


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma):
    x_val = sigma * -np.log(2.0 * erfinv(1 - q) ** 2) + mu
    return ppf_bounds_cont(x_val, q, -np.inf, np.inf)


@nb.njit(cache=True)
def nb_entropy(sigma):
    return 0.5 * (np.log(2 * np.pi * np.e * sigma**2))


@nb.njit(cache=True)
def nb_logpdf(x, mu, sigma):
    z_val = (x - mu) / sigma
    return -(1 / 2) * (z_val + np.exp(-z_val)) - np.log(sigma) - (1 / 2) * np.log(2 * np.pi)


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma):
    return -(nb_logpdf(x, mu, sigma)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from preliz.distributions.normal import Normal
from .distributions import Continuous
from ..internal.distribution_helper import eps, to_precision, from_precision, all_not_none
from ..internal.special import (
    beta,
    betainc,
    betaincinv,
    digamma,
    gammaln,
    erf,
    erfinv,
    ppf_bounds_cont,
)
from ..internal.optimization import optimize_ml


class StudentT(Continuous):
    r"""
    StudentT's distribution.

    Describes a normal variable whose precision is gamma distributed.

    The pdf of this distribution is

    .. math::

       f(x \mid \nu, \mu, \sigma) =
           \frac{\Gamma \left(\frac{\nu+1}{2} \right)} {\sqrt{\nu\pi}\
           \Gamma \left(\frac{\nu}{2} \right)} \left(1+\frac{x^2}{\nu} \right)^{-\frac{\nu+1}{2}}

    .. plot::
        :context: close-figs

        
        from preliz import StudentT, style
        style.use('preliz-doc')
        nus = [2., 5., 5.]
        mus = [0., 0.,  -4.]
        sigmas = [1., 1., 2.]
        for nu, mu, sigma in zip(nus, mus, sigmas):
            StudentT(nu, mu, sigma).plot_pdf(support=(-10,6))

    ========  ========================
    Support   :math:`x \in \mathbb{R}`
    Mean      :math:`\mu` for :math:`\nu > 1`, otherwise undefined
    Variance  :math:`\frac{\nu}{\nu-2}` for :math:`\nu > 2`,
              :math:`\infty` for :math:`1 < \nu \le 2`, otherwise undefined
    ========  ========================

    StudentT distribution has 2 alternative parameterization. In terms of nu, mu and
    sigma (standard deviation as nu increases) or nu, mu and lam (precision as nu increases).

    The link between the 2 alternatives is given by

    .. math::

        \lambda = \frac{1}{\sigma^2}

    Parameters
    ----------
    nu : float
        Degrees of freedom, also known as normality parameter (nu > 0).
    mu : float
        Location parameter.
    sigma : float
        Scale parameter (sigma > 0). Converges to the standard deviation as nu
        increases.
    lam : float
        Scale parameter (lam > 0). Converges to the precision as nu increases.
    """

    def __init__(self, nu=None, mu=None, sigma=None, lam=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self.params_support = ((eps, np.inf), (-np.inf, np.inf), (eps, np.inf))
        self._parametrization(nu, mu, sigma, lam)

    def _parametrization(self, nu=None, mu=None, sigma=None, lam=None):
        if all_not_none(sigma, lam):
            raise ValueError(
                "Incompatible parametrization. Either use nu, mu and sigma, or nu, mu and lam."
            )

        self.param_names = ("nu", "mu", "sigma")
        self.params_support = ((eps, np.inf), (-np.inf, np.inf), (eps, np.inf))

        if lam is not None:
            self.lam = lam
            sigma = from_precision(lam)
            self.param_names = ("nu", "mu", "lam")

        self.nu = nu
        self.mu = mu
        self.sigma = sigma

        if all_not_none(self.nu, self.mu, self.sigma):
            self._update(self.nu, self.mu, self.sigma)

    def _update(self, nu, mu, sigma):
        self.nu = np.float64(nu)
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.lam = to_precision(sigma)

        if self.param_names[2] == "sigma":
            self.params = (self.nu, self.mu, self.sigma)
        elif self.param_names[2] == "lam":
            self.params = (self.nu, self.mu, self.lam)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.nu, self.mu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.nu, self.mu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.nu, self.mu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.nu, self.mu, self.sigma)

    def entropy(self):
        return nb_entropy(self.nu, self.sigma)

    def mean(self):
        return self.mu

    def median(self):
        return self.mu

    def var(self):
        if self.nu > 2:
            return self.sigma**2 * self.nu / (self.nu - 2)
        elif self.nu > 1:
            return np.inf
        else:
            return np.nan

    def std(self):
        if self.nu > 2:
            return self.var() ** 0.5
        elif self.nu > 1:
            return np.inf
        else:
            return np.nan

    def skewness(self):
        if self.nu > 3:
            return 0
        else:
            return np.nan

    def kurtosis(self):
        if self.nu > 4:
            return 6 / (self.nu - 4)
        elif self.nu > 2:
            return np.inf
        else:
            return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return np.where(
            self.nu > 1e10,
            Normal(self.mu, self.sigma).rvs(size, random_state),
            random_state.standard_t(self.nu, size) * self.sigma + self.mu,
        )

    def _fit_moments(self, mean, sigma):
        # if nu is smaller than 2 the variance is not defined,
        # so if that happens we use 2.1 as an approximation
        nu = self.nu
        if nu is None:
            nu = 100
        elif nu <= 2:
            nu = 2.1
        else:
            sigma = sigma / (nu / (nu - 2)) ** 0.5

        self._update(nu, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, nu, mu, sigma):
    x = (x - mu) / sigma
    factor = 0.5 * betainc(0.5 * nu, 0.5, nu / (x**2 + nu))
    x_vals = np.where(x < 0, factor, 1 - factor)
    return np.where(nu > 1e10, 0.5 * (1 + erf((x - mu) / (sigma * 2**0.5))), x_vals)


@nb.njit(cache=True)
def nb_ppf(p, nu, mu, sigma):
    q = np.where(p < 0.5, p, 1 - p)
    x = betaincinv(0.5 * nu, 0.5, 2 * q)
    x = np.where(p < 0.5, -np.sqrt(nu * (1 - x) / x), np.sqrt(nu * (1 - x) / x))
    vals = np.where(nu > 1e10, mu + sigma * 2**0.5 * erfinv(2 * p - 1), mu + sigma * x)
    return ppf_bounds_cont(vals, p, -np.inf, np.inf)


@nb.vectorize(nopython=True, cache=True)
def nb_entropy(nu, sigma):
    if nu > 1e10:
        return 0.5 * (np.log(2 * np.pi * np.e * sigma**2))
    else:
        return (
            np.log(sigma)
            + 0.5 * (nu + 1) * (digamma(0.5 * (nu + 1)) - digamma(0.5 * nu))
            + np.log(np.sqrt(nu) * beta(0.5 * nu, 0.5))
        )


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, nu, mu, sigma):
    if nu > 1e10:
        return -np.log(sigma) - 0.5 * np.log(2 * np.pi) - 0.5 * ((x - mu) / sigma) ** 2
    else:
        return (
            gammaln((nu + 1) / 2)
            - gammaln(nu / 2)
            - 0.5 * np.log(nu * np.pi * sigma**2)
            - 0.5 * (nu + 1) * np.log(1 + ((x - mu) / sigma) ** 2 / nu)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, nu, mu, sigma):
    return -(nb_logpdf(x, nu, mu, sigma)).sum()
